import hashlib
import json
from abc import abstractmethod
from typing import (
    Any,
    Dict,
    List,
    Protocol,
    Tuple,
    runtime_checkable,
    TypedDict,
    Literal,
)
from urllib.parse import quote

from harambe.tracker import FileDataTracker
from harambe.types import URL, Context, Stage


ObservationTrigger = Literal[
    "on_save_data", "on_queue_url", "on_download", "on_paginate"
]


@runtime_checkable
class OutputObserver(Protocol):
    @abstractmethod
    async def on_save_data(self, data: Dict[str, Any]):
        raise NotImplementedError()

    @abstractmethod
    async def on_queue_url(self, url: URL, context: Dict[str, Any]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        raise NotImplementedError()

    @abstractmethod
    async def on_paginate(self, next_url: str) -> None:
        raise NotImplementedError()


class LoggingObserver(OutputObserver):
    async def on_save_data(self, data: Dict[str, Any]):
        print(data)

    async def on_queue_url(self, url: URL, context: Dict[str, Any]) -> None:
        print(f"Enqueuing: {url} with context {context}")

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        print(f"Downloading file: {filename}")  # TODO: use logger
        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass


class LocalStorageObserver(OutputObserver):
    def __init__(self, domain: str, stage: Stage):
        self._tracker = FileDataTracker(domain, stage)

    async def on_save_data(self, data: Dict[str, Any]) -> None:
        self._tracker.save_data(data)

    async def on_queue_url(self, url: URL, context: Dict[str, Any]) -> None:
        self._tracker.save_data({"url": url, "context": context})

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        data = {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }
        self._tracker.save_data(data)
        return data

    async def on_paginate(self, next_url: str) -> None:
        pass


class InMemoryObserver(OutputObserver):
    def __init__(self):
        self._data: List[Dict[str, Any]] = []
        self._urls: List[Tuple[URL, Context]] = []
        self._files: List[Tuple[str, bytes]] = []

    async def on_save_data(self, data: Dict[str, Any]) -> None:
        self._data.append(data)

    async def on_queue_url(self, url: URL, context: Dict[str, Any]) -> None:
        self._urls.append((url, context))

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        data = {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }
        self._files.append((filename, content))
        return data

    async def on_paginate(self, next_url: str) -> None:
        pass

    @property
    def data(self) -> List[Dict[str, Any]]:
        return self._data

    @property
    def urls(self) -> List[Tuple[URL, Context]]:
        return self._urls

    @property
    def files(self) -> List[Tuple[str, bytes]]:
        return self._files


class DuplicateHandler:
    def __init__(self):
        self._saved_data: set[bytes] = set()
        self.rows_on_page = 0
        self.previously_saved_rows_on_page = 0

    async def on_save_data(self, data: dict[str, Any]) -> bool:
        return self._add_data(data)

    async def on_queue_url(self, url: URL, context: dict[str, Any]) -> bool:
        return self._add_data(url)

    # noinspection PyTypeChecker
    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> bool:
        return self._add_data((download_url, filename))

    async def on_paginate(self, next_url: str) -> bool:
        if self.rows_on_page == self.previously_saved_rows_on_page:
            raise StopAsyncIteration()

        self.rows_on_page = 0
        self.previously_saved_rows_on_page = 0
        return False

    def _add_data(self, data: Any) -> bool:
        self.rows_on_page += 1

        hash_value = self.compute_hash(data)
        if hash_value in self._saved_data:
            self.previously_saved_rows_on_page += 1
            return True  # return True if data is duplicated
        else:
            self._saved_data.add(hash_value)
            return False

    @staticmethod
    def compute_hash(data: Any) -> bytes:
        if isinstance(data, dict):
            data = {k: v for k, v in data.items() if not k.startswith("__")}

        data_str = json.dumps(data, separators=(",", ":"), sort_keys=True)
        return hashlib.md5(data_str.encode()).digest()


class DownloadMeta(TypedDict):
    url: str
    filename: str



from pydantic.functional_validators import AfterValidator
from typing import List
from typing_extensions import Annotated

from harambe.types import Enum


class ParserTypeEnum:
    def __new__(self, variants: List[Enum]):
        return Annotated[Enum, AfterValidator(self.validate_type(variants))]

    def validate_type(variants: List[Enum]):
        def _validate_type(value: Enum) -> Enum:
            # Check if the value exists among variants
            if value not in variants:
                raise ValueError(
                    f'Value "{value}" doesn\'t match any of the variants ({variants})'
                )
            return value

        return _validate_type



from .core import SDK, AsyncScraper, PAGE_PDF_FILENAME
from .parser.schemas import Schemas
from .types import AsyncScraperType, ScrapeResult
from .utils import PlaywrightUtils

__all__ = [
    "ScrapeResult",
    "Schemas",
    "SDK",
    "PlaywrightUtils",
    "AsyncScraperType",
    "AsyncScraper",
    "PAGE_PDF_FILENAME",
]



from contextlib import asynccontextmanager
from typing import AsyncGenerator
from urllib.parse import urlparse

from playwright.async_api import async_playwright, Page, ProxySettings
from playwright_stealth import stealth_async

from harambe.handlers import UnnecessaryResourceHandler


def proxy_from_url(url: str) -> ProxySettings:
    parsed = urlparse(url, allow_fragments=False)

    if not parsed.hostname:
        raise ValueError(f"Invalid proxy URL: {url}")

    proxy: ProxySettings = {
        "server": parsed.hostname,
        "username": parsed.username,
        "password": parsed.password,
    }

    if parsed.port:
        proxy["server"] += f":{parsed.port}"

    return proxy


@asynccontextmanager
async def playwright_harness(
    headless: bool,
    cdp_endpoint: str | None,
    proxy: str | None = None,
) -> AsyncGenerator[Page, None]:
    """
    Context manager for Playwright. Starts a new browser, context, and page, and closes them when done.
    Also does some basic setup like setting the viewport, user agent, ignoring HTTPS errors, creation of HAR file, and stealth.

    :param headless: launch browser in headless mode
    :param cdp_endpoint: Chrome DevTools Protocol endpoint to connect to (if using a remote browser)
    :return: async generator yielding a Playwright page
    """

    async with async_playwright() as p:
        browser = await (
            p.chromium.connect_over_cdp(endpoint_url=cdp_endpoint)
            if cdp_endpoint
            else p.chromium.launch(headless=headless)
        )

        ctx = await browser.new_context(
            viewport={"width": 1280, "height": 1024},
            ignore_https_errors=True,
            user_agent="Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
            proxy=proxy_from_url(proxy) if proxy else None,
        )

        ctx.set_default_timeout(60000)
        await ctx.route("**/*", UnnecessaryResourceHandler().handle)

        page = await ctx.new_page()
        await stealth_async(page)

        try:
            yield page
        finally:
            await ctx.close()
            await browser.close()



from pydantic.functional_validators import AfterValidator
from typing import Optional
from typing_extensions import Annotated
import urllib.parse
from urllib.parse import urljoin, urlparse

from harambe.types import URL

allowed_url_schemes = [
    "data",
    "file",
    "http",
    "https",
    "mailto",
    "s3",
]


class ParserTypeUrl:
    def __new__(self, base_url: Optional[URL] = None):
        return Annotated[URL, AfterValidator(self.validate_type(base_url))]

    def validate_type(base_url: Optional[URL]):
        def _validate_type(url: URL) -> str:
            # Transform relative URLs into absolute using base_url
            if base_url is not None:
                parsed_base_url = urlparse(base_url, allow_fragments=False)
                url = urljoin(parsed_base_url.geturl(), url)

            # Parse the URL
            try:
                parsed_url = urllib.parse.urlparse(url)
            except ValueError as e:
                raise ValueError(f"Unable to parse URL: {url}", e)

            # Check if the scheme is allowed
            if parsed_url.scheme not in allowed_url_schemes:
                raise ValueError(f"Invalid URL: {url}")

            return url

        return _validate_type



from abc import ABC, abstractmethod
from pydantic import BaseModel, create_model, Extra, Field, NameEmail, ValidationError
from typing import Any, Dict, List, Optional, Type

from harambe.parser.type_date import ParserTypeDate
from harambe.parser.type_enum import ParserTypeEnum
from harambe.parser.type_phone_number import ParserTypePhoneNumber
from harambe.parser.type_url import ParserTypeUrl
from harambe.types import Schema, URL

OBJECT_TYPE = "object"
LIST_TYPE = "array"


class SchemaParser(ABC):
    """
    Interface for schema parsers
    """

    @abstractmethod
    def validate(self, data: Dict[str, Any]) -> None:
        pass


class SchemaValidationError(Exception):
    def __init__(self, schema, data, message):
        super().__init__(
            "Data {data} does not match schema {schema}. {message}".format(
                data=data, schema=schema, message=message
            )
        )


class PydanticSchemaParser(SchemaParser):
    """
    A schema parser that uses Pydantic models to validate data against a JSON schema
    """

    def __init__(self, schema: Schema):
        self.schema = schema
        self.base_url = None

    def validate(self, data: Dict[str, Any], base_url: URL) -> None:
        self.base_url = base_url
        self.field_types = self._get_field_types()
        self.model = self._schema_to_pydantic_model(self.schema)

        try:
            self.model(**data)
        except ValidationError as validation_error:
            raise SchemaValidationError(
                data=data, schema=self.schema, message=validation_error
            )

    def _get_field_types(self) -> Dict[str, Type]:
        return {
            "string": str,
            "str": str,
            "boolean": bool,
            "bool": bool,
            "integer": int,
            "int": int,
            "number": float,
            "float": float,
            "double": float,
            "email": NameEmail,
            "enum": ParserTypeEnum,
            LIST_TYPE: List,
            OBJECT_TYPE: Dict[str, Any],
            "datetime": ParserTypeDate(),
            "phone_number": ParserTypePhoneNumber(),
            "url": ParserTypeUrl(base_url=self.base_url),
        }

    def _items_schema_to_python_type(
        self, items_info: Schema, model_name: str = "DynamicModelItem"
    ) -> Type:
        """
        Convert a JSON schema's items property to a Python type
        """
        item_type = items_info.get("type")

        if item_type == OBJECT_TYPE:
            python_type = self._schema_to_pydantic_model(
                items_info.get("properties", {}),
                model_name=f"{model_name}Object",
            )
        elif item_type == LIST_TYPE:
            # Lists can't be null
            python_type = List[
                self._items_schema_to_python_type(
                    items_info.get("items", {}),
                    model_name=f"{model_name}Item",
                )
            ]
        elif item_type == "enum":
            # Every enum has its own unique variants
            python_type = self._get_type(item_type)(items_info["variants"])
        else:
            # Non complex types aren't optional when they're within a list
            python_type = self._get_type(item_type)

        return python_type

    def _schema_to_pydantic_model(
        self, schema: Schema, model_name: str = "DynamicModel"
    ) -> Type[BaseModel]:
        """
        Convert a JSON schema to a Pydantic model dynamically. All fields are optional
        """
        fields = {}

        for field_name, field_info in schema.items():
            field_description = field_info.get("description", None)
            field_type = field_info.get("type")

            if field_type == OBJECT_TYPE:
                python_type = self._schema_to_pydantic_model(
                    field_info.get("properties", {}),
                    model_name=f"{model_name}{field_name.capitalize()}",
                )
            elif field_type == LIST_TYPE:
                # Lists can't be null
                python_type = List[
                    self._items_schema_to_python_type(
                        field_info.get("items", {}),
                        model_name=f"{model_name}Item",
                    )
                ]
            elif field_type == "enum":
                # Every enum has its own unique variants
                python_type = self._get_type(field_type)(field_info["variants"])
            else:
                # Non complex types should be optional
                python_type = Optional[self._get_type(field_type)]

            fields[field_name] = (
                python_type,
                Field(..., description=field_description),
            )

        # Disallow additional fields
        config = {"extra": Extra.forbid}

        return create_model(model_name, __config__=config, **fields)

    def _get_type(self, field: str) -> Type:
        field_type = self.field_types.get(field)
        if not field_type:
            raise ValueError(f"Unsupported field type: {field}")
        return field_type



from pydantic.functional_validators import AfterValidator
import re
from typing_extensions import Annotated

phone_number_formats = [
    r"^\d{3,11}$",  # 911 & 11111111111
    r"^(\(?\+\d{1,2}\)?\s)?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}$",  # +1 (628) 555-3456 & (+1) 415-155-1555
    r"^\(\+\d{1,3}\)\s\d{10}(\s\(Extension:\s\d{1,4}\))?$",  # (+4) 1111111111 (Extension: 323)
    r"^(\(?\+\d{1,2}\)?\s)?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}(,\s\(?ext\.\s\d{1,4}\)?)?$",  # 206-555-7115 & 206-555-7115, ext. 239
]


class ParserTypePhoneNumber:
    def __new__(self):
        return Annotated[str, AfterValidator(self.validate_type)]

    def validate_type(number: str) -> str:
        # Trim whitespaces
        number = number.strip()
        # Attempt to parse phone number
        for phone_number_format in phone_number_formats:
            regex = re.compile(phone_number_format)
            if regex.match(number):
                return number
        raise ValueError(f"Unable to parse input as phone number: {number}")




import asyncio
from typing import Any

from playwright.async_api import Page

from harambe import SDK


async def scrape(sdk: SDK, url: str, *args: Any, **kwargs: Any) -> None:
    page: Page = sdk.page
    await asyncio.sleep(20)
    await page.wait_for_selector("div.table_wrapper table tbody tr td a")
    links = await page.query_selector_all("div.table_wrapper table tbody tr td a")
    for link in links:
        href = await link.get_attribute("href")
        await sdk.enqueue(href)

if __name__ == "__main__":
    asyncio.run(SDK.run(scrape, "https://account.bonfirehub.com/login", {}))



from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

import dateparser


class ParserTypeDate:
    def __new__(self):
        return Annotated[str, AfterValidator(self.validate_type)]

    def validate_type(date: str):
        if not isinstance(date, str):
            raise ValueError("Wrong input type")

        # Trim whitespaces
        date = date.strip()

        # Make sure it's not empty string
        if len(date) == 0:
            raise ValueError("Empty input")

        # Attempt to parse date string
        try:
            dateparser.parse(date)
            return date
        except ValueError:
            pass

        raise ValueError(f"Unable to parse input as date: {date}")



from typing import Any, Awaitable, Callable, Dict, Literal

Enum = str
URL = str
ScrapeResult = Dict[str, Any]
Context = Dict[str, Any]
Stage = Literal["category", "listing", "detail"]
AsyncScraperType = Callable[["SDK", URL, Context], Awaitable[None]]  # noqa: F821
SetupType = Callable[["SDK"], Awaitable[None]]  # noqa: F821
Schema = Dict[str, Any]
